@@ -19,6 +19,7 @@
 # ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 # CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 # SOFTWARE.
+import asyncio
 import json
 import locale
 import traceback
@@ -26,14 +27,14 @@ import subprocess
 import sys
 import os
 
-import requests
-
 from .version import ELECTRUM_VERSION
-from .import constants
+from . import constants
 from .i18n import _
+from .util import make_aiohttp_session
+from .logging import describe_os_version, Logger
 
 
-class BaseCrashReporter(object):
+class BaseCrashReporter(Logger):
     report_server = "https://crashhub.electrum.org"
     config_key = "show_crash_reporter"
     issue_template = """<h2>Traceback</h2>
@@ -58,18 +59,25 @@ class BaseCrashReporter(object):
     ASK_CONFIRM_SEND = _("Do you want to send this report?")
 
     def __init__(self, exctype, value, tb):
+        Logger.__init__(self)
         self.exc_args = (exctype, value, tb)
 
-    def send_report(self, endpoint="/crash"):
+    def send_report(self, asyncio_loop, proxy, endpoint="/crash", *, timeout=None):
         if constants.net.GENESIS[-4:] not in ["4943", "e26f"] and ".electrum.org" in BaseCrashReporter.report_server:
             # Gah! Some kind of altcoin wants to send us crash reports.
             raise Exception(_("Missing report URL."))
         report = self.get_traceback_info()
         report.update(self.get_additional_info())
         report = json.dumps(report)
-        response = requests.post(BaseCrashReporter.report_server + endpoint, data=report)
+        coro = self.do_post(proxy, BaseCrashReporter.report_server + endpoint, data=report)
+        response = asyncio.run_coroutine_threadsafe(coro, asyncio_loop).result(timeout)
         return response
 
+    async def do_post(self, proxy, url, data):
+        async with make_aiohttp_session(proxy) as session:
+            async with session.post(url, data=data) as resp:
+                return await resp.text()
+
     def get_traceback_info(self):
         exc_string = str(self.exc_args[1])
         stack = traceback.extract_tb(self.exc_args[2])
@@ -89,7 +97,7 @@ class BaseCrashReporter(object):
         args = {
             "app_version": ELECTRUM_VERSION,
             "python_version": sys.version,
-            "os": self.get_os_version(),
+            "os": describe_os_version(),
             "wallet_type": "unknown",
             "locale": locale.getdefaultlocale()[0] or "?",
             "description": self.get_user_description()
@@ -123,6 +131,3 @@ class BaseCrashReporter(object):
 
     def get_wallet_type(self):
         raise NotImplementedError
-
-    def get_os_version(self):
-        raise NotImplementedError 
